=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed August 2009 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   GhsotCompAlgo.rb
# Original Date	:   10 Aug 2009 - version 1.0
# Description	:   Algorithms for the Ghost Component script
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module GhostComp

#---------------------------------------------------------------------------------------------------------------------------
#Constants for GhostComp  Modules (do not translate here, use Translation Dialog Box instead)	
#---------------------------------------------------------------------------------------------------------------------------

T6[:TITLE_GhostComponent] = "Ghost Components"

T6[:WARNING_OpenGhost] = "You should not Edit a GHOST component"		
T6[:WARNING_DeleteGhost] = "Do you confirm the deletion of ghosts for %1 components"
T6[:WARNING_SelectAll] = "You cannot generate ghosts for the whole model"
T6[:WARNING_NoCompSelected] = "There are NO components in the selection"
T6[:WARNING_PurgeUnused] = "Perform a Purge Unused for components in the model before clean up"

T6[:BUTTON_Edit] = "Edit anyway"
T6[:BUTTON_Close] = "Close Component"
T6[:BUTTON_Switch] = "Switch to Real"

T6[:DESC_GhostName] = "Ghost definition for %1"
T6[:OPS_CreateGhost] = "Generate Ghost component for selection"	
T6[:OPS_ToGhost] = "Switch to Ghost"	
T6[:OPS_ToReal] = "Switch to Real"	
T6[:OPS_ColorGhost] = "Ghost color to %1"	

#--------------------------------------------------------
# Constant for dictionary (do not translate)
#--------------------------------------------------------

GHC__Dico = "__GhostComp__"
GHC__Type = "__Type__"
GHC__Name = "__NAME__"
GHC__PREFIX = "zzGHC_"

GHC__LAYER = "__GHOSTCOMP__LAYER__"
GHC__MAT = "__GHOSTCOMP__MATERIAL__"

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# Class GhostCompAlgo: main algorithm for the plugin
#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
	
class GhostCompAlgo	

#Pair Main / GHost component definition
GHC_Pair = Struct.new "GHC_Pair", :name, :main_cdef, :ghost_cdef

@@model = nil
@@hsh_pairs = nil
@@hsh_maindef = {}
@@time_last_ops = nil
@@double_click_ops = nil

def GhostCompAlgo.launcher(ops_symb, ops_symb_double=nil, selection=nil)
	#Detecting a double click
	t0 = Time.now.to_f
	double = true
	if @@time_last_ops && (t0 - @@time_last_ops < 0.5) && @@double_click_ops == nil
		@@double_click_ops = ops_symb_double
	else	
		@@double_click_ops = nil
	end	
	@@time_last_ops = t0
	algo = GhostCompAlgo.new
	algo.launch_operation ops_symb, nil, ops_symb_double
end

#Initialization
def initialize(ops_symb=nil, selection=nil, ops_symb_double=nil)
	#check if new model
	@model = Sketchup.active_model
	GhostCompAlgo.reset if @@model == nil || @model.guid != @@model.guid
	
	#Context initialization
	@entities = @model.entities
	@active_entities = @model.active_entities
	@definitions = @model.definitions
	@layers = @model.layers
	@materials = @model.materials
	@selection = @model.selection
	@tr_id = Geom::Transformation.new
	@operation = false
	@mode_ghost = GhostComp.get_option
	@timer_commitment = false
	@cube_face = [[0, 1, 3, 2], [0, 1, 5, 4], [1, 3, 7, 5], [4, 5, 7, 6], [0, 4, 6, 2], [3, 2, 6, 7]]
	@simplify_face_mini = MYDEFPARAM[:GHC_DEFAULT_NbSimple] - 1
	@touched_cdef = []	
end	

#Check if the model has changed (new file, open file, revert)
def GhostCompAlgo.reset
	#installing the selection observer
	@@hsh_pairs = nil
	@@model = Sketchup.active_model
	if MYDEFPARAM[:GHC_DEFAULT_UseObserver]
		obj = (SU_MAJOR_VERSION < 7) ? @@model.selection : @@model
		obj.add_observer GhostSelectionSpy.new
	end
end

#Launch the requested operation
def launch_operation(ops_symb, selection=nil, ops_symb_double=nil, nomessage=false)
	#Evaluating the selection
	selection = @selection.to_a unless selection
	if selection.length == 0
		if ops_symb == :OPS_CreateGhost
			UI.messagebox T6[:WARNING_SelectAll]
			return false
		end
		selection = @active_entities.to_a
	end	
	sall = (selection.length == @active_entities.length)
	selection = selection.find_all { |e| e.instance_of?(Sketchup::ComponentInstance) }
	if selection.length == 0
		UI.messagebox T6[:WARNING_NoCompSelected] unless nomessage
		return false
	end	

	#Finding the components
	recurse = (ops_symb == :OPS_CreateGhost && @mode_ghost != :GHC_OPT_BOX_) || 
	          ops_symb == :OPS_DeleteGhost || ops_symb == :OPS_ToRealAll
	hshdef = find_components selection, recurse
	return false if hshdef.length == 0
	lstdef = hshdef.values
	
	#Confirmation for ghost deletion
	if ops_symb == :OPS_DeleteGhost
		snb = (sall) ? "ALL" : lstdef.length.to_s
		return false if UI.messagebox(T6[:WARNING_DeleteGhost, snb], MB_YESNO) != 6
	end

	#Processing the operation
	if ops_symb_double
		UI.start_timer(0.1) { process_operation ops_symb, selection, lstdef, ops_symb_double }
	else
		process_operation ops_symb, selection, lstdef
	end
	true	
end

def process_operation(ops_symb, selection, lstdef, ops_symb_double=nil)
	#Initiating the operation
	start_operation ops_symb
	
	build_all_pairs
	@material_for_ghosts = get_material_for_ghosts
	@layer_for_ghosts = get_layer_for_ghosts
	
	#Performing the operation
	case ops_symb

	when :OPS_CreateGhost
		do_create_ghosts lstdef, selection
	
	when :OPS_DeleteGhost
		do_delete_ghosts lstdef

	when :OPS_ToGhost
		lstdef.each { |lst| to_ghost_component lst }

	when :OPS_ToReal, :OPS_ToRealAll
		lstdef.each { |lst| to_real_component lst }
		
	end	
	
	#Committing the operation, unless deferred in a timer
	commit_operation unless @timer_commitment
	
	true
end

#Find component definitions and corresponding instances in the given selection
def find_components(selection, recurse=false, hshdef=nil)
	hshdef = {} unless hshdef
	gh_layer = @layers[GHC__LAYER]
	
	selection.each do |entity|
		#For component instance, check if pure container
		if entity.instance_of? Sketchup::ComponentInstance
			next if gh_layer && entity.layer == gh_layer
			dc = G6.is_dynamic_component?(entity)
			@dyncomp = true if dc
			cdef = entity.definition
			hshdef[cdef.entityID] = [cdef] unless hshdef[cdef.entityID]
			hshdef[cdef.entityID].push entity
			find_components cdef.entities, recurse, hshdef if recurse
			#hshdef[cdef.entityID].push entity
			
		#Group considered only of pure container	
		elsif entity.instance_of?(Sketchup::Group) && pure_container?(entity)
			find_components entity.entities, recurse, hshdef
		end	
	end
	hshdef
end

#Check if Component or Group is a pure container
def pure_container?(cdef)
	(cdef.entities.find { |e| !e.instance_of?(Sketchup::ComponentInstance) && !e.instance_of?(Sketchup::Group) } ) ? false : true
end

#Short cut for switching to Ghost, without start operation
def switch_to_ghost(selection)
	launch_operation :OPS_ToGhost, selection
end

#Short cut for switching to Real, without start operation
def switch_to_real(selection)
	launch_operation :OPS_ToReal, selection
end

#-------------------------------------------------------------------
# Managing SU operations operations
#-------------------------------------------------------------------

def start_operation(ops_symb, param=nil)
	return if @operation
	G6.start_operation @model, T6[ops_symb, param]
	@operation = true
end

def commit_operation
	return unless @operation
	@model.commit_operation
	@operation = false
	@touched_cdef.each { |cdef| cdef.refresh_thumbnail } if SU_MAJOR_VERSION >= 7
	
	#Chaining for possible double click
	if @@double_click_ops
		ops = @@double_click_ops
		@@double_click_ops = nil
		launch_operation ops
	end	
		
end

def abort_operation
	return unless @operation
	@model.abort_operation
	@operation = false
end

#-------------------------------------------------------------------
# Managing layers and material
#-------------------------------------------------------------------

#Set or return the ghost comp layer
def get_layer_for_ghosts
	ly = @layers[GHC__LAYER]
	return ly if ly
	ly = @layers.add GHC__LAYER
	ly.visible = false
	ly
end

#Return the material for ghost
def get_material_for_ghosts
	mat0 = @materials[GHC__MAT]
	mat = (mat0) ? mat0 : @materials.add(GHC__MAT)
	mat.color = MYDEFPARAM[:GHC_COLOR_RenderGhost]
	mat	
end
	
#Update the Ghost color (from the Default parameter box)	
def update_ghost_color
	mat = @materials[GHC__MAT]
	return unless mat
	color = MYDEFPARAM[:GHC_COLOR_RenderGhost]
	start_operation :OPS_ColorGhost, color
	mat.color = color
	refresh_thumbnails
	commit_operation
end	

#Launch a clean up operation from the icon or menu command
def launch_cleanup
	status = UI.messagebox T6[:WARNING_PurgeUnused], MB_YESNO
	@definitions.purge_unused if status == 6
		
	start_operation :GHC_CLEANUP_Menu
	
	@@hsh_pairs = nil
	build_all_pairs
	
	@definitions.purge_unused if status == 6
	
	commit_operation
end
#-------------------------------------------------------------------
# Top level routines to perform operations
#-------------------------------------------------------------------

#Top level routine for creating ghosts
def do_create_ghosts(lstdef, selection)
	ng = 0
	pbar = Traductor::ProgressionBar.new lstdef.length, "Step:"
	lstdef.each do |lst| 
		ng += 1 if generate_ghost lst.first
		pbar.countage
	end
	populate_ghosts
	Sketchup.set_status_text "DONE #{ng}", SB_VCB_LABEL
	Sketchup.set_status_text ""
	if ng > 0
		return switch_to_ghost(selection) if MYDEFPARAM[:GHC_DEFAULT_AutoSwitch]
		animate selection if ng <= 15
	end	
end

#animation after generating ghosts
def animate(selection)
	#quick animation
	@timer_commitment = true
	UI.start_timer(0.1) do
		switch_to_ghost(selection)
		UI.start_timer(0.1) { sleep 0.1 ; switch_to_real(selection) ; commit_operation }
	end	
end

#Create a ghsot component from a given definition	
def generate_ghost(main_cdef)
	#prevent to make a ghost from a ghost
	return nil if is_ghost?(main_cdef)
	
	#Remove previous ghost component if any and substitute temporarily components
	lst_using = []
	lgh = @@hsh_maindef[main_cdef.entityID]
	if lgh
		lgh = lgh.find_all { |gdef| gdef.valid? }
		lcomp = []
		lgh.each { |gdef| lcomp += gdef.instances.find_all { |e| e.layer == @layer_for_ghosts } }
		@entities.erase_entities lcomp if lcomp.length > 0
		lgh.each do |gdef|  
			lcompold = gdef.instances.to_a
			lst_using += lcompold
			to_real_component [gdef] + lcompold
		end	
	end 
	
	#Create a ghost definition of the main definition if there was no previous ghost
	gh_comp = create_ghost main_cdef 
	return nil unless gh_comp
	gh_cdef = gh_comp.definition
	
	#Registering the definitions
	name = register_pair main_cdef, gh_cdef
	
	#Put the ghost component of the Ghost layer
	mark_component gh_comp, name, "Ghost"
	
	#Create an instance of main definition if does not exist (to prevent disparition via purge unused)
	main_comp = main_cdef.instances.find { |e| e.layer == @layer_for_ghosts }	
	main_comp = @entities.add_instance main_cdef, @tr_id unless main_comp
	mark_component main_comp, name, "Main" if main_comp
	
	#turning back to ghosts components using the old ghost definition and Removing the previous definitions if any
	if lgh
		to_ghost_component [main_cdef] + lst_using if lst_using.length > 0
		lgh.each { |gdef| gdef.entities.erase_entities gdef.entities.to_a if gdef.valid? }
	end
	
	#returning the ghost definition
	@touched_cdef.push gh_cdef
	
	gh_cdef
end
	
#Swicth model to show Ghost components	
def to_ghost_component(lst)
	#Getting the main definition
	main_cdef = lst.first
	gh_cdef = ghost_from_main main_cdef, true
	return unless gh_cdef && gh_cdef.valid?
		
	#Transfering the definitions
	lst[1..-1].each do |comp|
		if comp.valid? && comp.layer != @layer_for_ghosts && is_main?(comp.definition)
			comp.definition = gh_cdef
		end	
	end		
end

#Swicth model to show Real components	
def to_real_component(lst)
	#Getting the main definition
	gh_cdef = lst.first
	main_cdef = main_from_ghost gh_cdef
	return unless main_cdef && main_cdef.valid?
		
	#Switching the component definition (and material due to a bug in SU7)
	lst[1..-1].each do |comp|
		if comp.valid? && comp.layer != @layer_for_ghosts && is_ghost?(comp.definition)
			comp.definition = main_cdef
		end	
	end		
end

#Launcher for the Ghost Removal 
def do_delete_ghosts(lstdef)

	#Turning ghost into Real components when applicable
	lst_realdef = []
	lst_ghdef = []
	lstdef.each do |lst|
		cdef = lst.first
		if is_ghost?(cdef)
			gh_cdef = cdef
			main_cdef = main_from_ghost gh_cdef
		elsif is_main?(cdef)
			gh_cdef = ghost_from_main(cdef)
			main_cdef = cdef
		else
			next
		end	
		lst_ghdef.push gh_cdef
		lst_realdef.push main_cdef, gh_cdef
		next unless gh_cdef
		instances = gh_cdef.instances
		next if instances.length == 0
		to_real_component([gh_cdef] + instances)
	end

	#Erasing the shadow components on the Ghost layer
	lst_realdef.each do |cdef|
		next unless cdef && cdef.valid?
		lcomp = cdef.instances.find_all { |e| e.layer == @layer_for_ghosts }
		@entities.erase_entities lcomp
	end
	
	#Deleting the ghost definitions
	lst_ghdef.each do |cdef|
		next unless cdef.valid?
		cdef.entities.erase_entities cdef.entities.to_a
	end
end

#-------------------------------------------------------------
# Methods to create a ghost component
#-------------------------------------------------------------

#Main method to create ghost components
def create_ghost(main_cdef)
	case @mode_ghost
	
	when :GHC_OPT_SIMPLE_
		gh_comp = create_ghost_copy main_cdef
		entities = gh_comp.definition.entities
		simplify_ghost entities
		colorize_ghost entities

	when :GHC_OPT_BOX_
		gh_comp = create_ghost_copy main_cdef
		gh_cdef = gh_comp.definition
		simple_boxify_ghost gh_cdef.entities

	when :GHC_OPT_MULTIBOX_
		gh_comp = create_ghost_copy main_cdef
		gh_cdef = gh_comp.definition
		multi_boxify_ghost gh_cdef.entities
	
	when :GHC_OPT_WIREFRAME_
		gh_comp = create_ghost_copy main_cdef
		gh_cdef = gh_comp.definition
		wireframe_ghost gh_cdef.entities
	
	else
		gh_comp = create_ghost_copy main_cdef
		colorize_ghost gh_comp.definition.entities
	
	end
	gh_comp
end	

#Create a unique instance of a definition
def create_ghost_copy(main_cdef)
	gh_comp = @entities.add_instance main_cdef, @tr_id
	gh_comp.make_unique
	avoid_problem_with_dc gh_comp
	gh_comp
end	

#Colorize a definition
def colorize_ghost(entities)
	ghmat = @material_for_ghosts
	entities.each do |e|
		if e.instance_of? Sketchup::Face
			e.material = e.back_material = ghmat
		elsif e.instance_of? Sketchup::Group
			e.make_unique #if SU_MAJOR_VERSION < 7
			colorize_ghost e.entities
		end
	end	
end

#boxify a component definition
def simple_boxify_ghost(entities)
	mat = @material_for_ghosts
	bbgeom = Geom::BoundingBox.new
	entities.each do |e|
		bbgeom = bbgeom.add e.bounds if e.bounds
	end	
	entities.erase_entities entities.to_a
	box_from_bounding_box entities, bbgeom, mat
end

#boxify a component definition
def multi_boxify_ghost(entities, within_group=false)
	lgeom = []
	lgroup = []
	lcomp = []
	entities.each do |e|
		if e.instance_of? Sketchup::Group
			lgroup.push e
		elsif e.instance_of? Sketchup::ComponentInstance
			lgeom.push e if within_group
			#lcomp.push e if within_group
		else
			lgeom.push e
		end	
	end
	
	#replacing all geometry by their bounding box
	if lgeom.length > 0
		mat = ghmat = @material_for_ghosts
		bbgeom = Geom::BoundingBox.new
		lgeom.each do |e|
			bbgeom = bbgeom.add e.bounds if e.valid? && e.bounds
		end	
		entities.erase_entities lgeom
		box_from_bounding_box entities, bbgeom, mat
	end	
	
	#Handling the groups recursively
	lgroup.each do |grp|
		next unless grp.valid?
		grp = grp.make_unique
		multi_boxify_ghost grp.entities, true
	end
	
	#Handling components embedded within groups recursively
	lcomp.each do |comp|
		next unless comp.valid?
		#comp.make_unique
		#multi_boxify_ghost comp.definition.entities, true
	end
	
end

#Create a box from a bounding box in the given entities
def box_from_bounding_box(entities, bb, mat)
	cr = []
	for i in 0..7 
		cr[i] = bb.corner i
	end	
	
	@cube_face.each do |ls|
		begin
			face = entities.add_face ls.collect { |i| cr[i] }
			face.material = face.back_material = mat
		rescue
		end	
	end	
end

#Wireframe a definition by keeping only edges
def wireframe_ghost(entities, within_group=false)
	lstdel = []
	lstedges = []
	entities.each do |e|
		if e.instance_of? Sketchup::Group
			e.make_unique
			wireframe_ghost e.entities, true
		elsif e.instance_of?(Sketchup::ComponentInstance)
			if within_group
				e.make_unique
				wireframe_ghost e.definition.entities, true
			end	
		elsif e.instance_of?(Sketchup::Edge)
			lstedges.push e
		else
			lstdel.push e
		end
	end	
	
	lstedges.each { |e| e.smooth = e.soft = false }
	entities.erase_entities lstdel if lstdel.length > 0
end

#Simplify a definition by keeping a limited number of faces
def simplify_ghost(entities, lst_entities=nil, within_group=nil)
	hshface = {}
	lst_entities = entities.to_a unless lst_entities
	lcomp = []
	lgroup = []
	
	#Exploding the entities recurvively
	lst_entities.each do |e|
		if e.instance_of? Sketchup::Face
			hshface[e.entityID] = [e, e.area] unless hshface[e.entityID]
		elsif e.instance_of? Sketchup::Group
			lgroup.push e
		elsif e.instance_of? Sketchup::ComponentInstance
			if within_group
				lcomp.push e
			end	
		end
	end	
	
	#Sorting the faces by area
	lkeep = hshface.values.sort { |a, b| a[1] <=> b[1] }
	lkeep = lkeep.reverse[0..@simplify_face_mini]
	lkeep = lkeep.collect { |ll| ll[0] }
	hshedge = {}
	lkeep.each do |face|
		face.edges.each { |edge| hshedge[edge.entityID] = edge }
	end	
	lkeep += hshedge.values
	
	#Deleting all elements except the faces and edges
	ldel = lst_entities.find_all { |e| !e.instance_of?(Sketchup::Group) && !e.instance_of?(Sketchup::ComponentInstance) }
	ldel = ldel - lkeep
	entities.erase_entities ldel
	
	#Handling the components
	lcomp.each do |comp|
		ent = comp.explode
		simplify_ghost entities, ent, within_group
	end

	#Handling the groups
	lgroup.each do |grp|
		grp.make_unique
		simplify_ghost grp.entities, nil, grp
	end
	
end

#--------------------------------------------------------
# Utility functions for Pair management
#--------------------------------------------------------

def refresh_thumbnails
	return unless SU_MAJOR_VERSION >= 7
	build_all_pairs
	@@hsh_pairs.each do |key, pair|
		gdef = pair.ghost_cdef
		gdef.refresh_thumbnail if gdef && gdef.valid?
	end
end

#Code necessary to bootstrap the Dynamic Component - Don't ask me why!	
def avoid_problem_with_dc(comp=nil)
	return if SU_MAJOR_VERSION < 7
	begin
		a = $dc_observers.get_latest_class
		#a.redraw if a && comp
	rescue
	end	
	sleep 0.1 if @dyncomp
	#@dyncomp = false
	@last_dc = a
end
	
#--------------------------------------------------------
# Managing the pairs Main / Ghost
#--------------------------------------------------------
		
#Create a pair Main / Ghost
def create_pair(name, main_cdef, ghost_cdef)
	name = Time.now.to_f.to_s unless name
	pair = @@hsh_pairs[name]
	unless pair
		pair = @@hsh_pairs[name] = GHC_Pair.new
		pair.name = name
	end	
	pair.main_cdef = main_cdef if main_cdef
	pair.ghost_cdef = ghost_cdef if ghost_cdef
	pair
end

#Get the pair corresponding to its name
def get_pair(name)
	return nil unless name
	build_all_pairs
	@@hsh_pairs[name]
end

#Build pairs from the current model by scanning all definitions - Done only when model is changed
def build_all_pairs(forced=false)
	return if @@hsh_pairs && !forced
	@@hsh_pairs = {}
	@@hsh_maindef = {}
	
	gh_layer = @layers[GHC__LAYER]
	return unless gh_layer
	lst_comp = []
	@definitions.each do |cdef|
		lcomp = cdef.instances.find_all { |e| e.layer == gh_layer }
		lst_comp += lcomp
		lcomp.each do |comp|
			name = comp.get_attribute GHC__Dico, GHC__Name
			next unless name
			if comp.get_attribute(GHC__Dico, GHC__Type) == "Main"
				create_pair name, cdef, nil
			elsif comp.get_attribute(GHC__Dico, GHC__Type) == "Ghost"
				create_pair name, nil, cdef
			end	
		end	
	end	

	#Clean up only when called first time
	clean_up_pairs(lst_comp, gh_layer) unless forced
	
	#keeping only valid pairs
	@@hsh_pairs.delete_if do |key, pair|
		!(pair.main_cdef && pair.main_cdef.valid? && pair.ghost_cdef && pair.ghost_cdef.valid?)
	end
	
	#Building the hash for main definition (to manage undo)
	@@hsh_pairs.each do |key, pair|
		associate_ghost_to_main pair.main_cdef, pair.ghost_cdef
	end
end

#Clean up copies on the Ghost layer and erase definitions not used
def clean_up_pairs(lst_comp, gh_layer)
	lst_gcomp = []
	ldel_comp = []
	hdel_cdef = {}
	hkeep_cdef = {}
	
	#Filtering the components on Ghost layer concrened by GhostComp (we never know!)
	lst_comp.find_all { |comp| lst_gcomp.push comp if comp.valid? && comp.get_attribute(GHC__Dico, GHC__Name) }
	
	#Finding out if these component have instance in the model, outside of the Ghost layer
	lst_gcomp.each do |comp|
		cdef = comp.definition
		next if hkeep_cdef[cdef.entityID]
		lst = G6.find_top_comp_where_used cdef
		hkeep_cdef[cdef.entityID] = false
		lst.each do |c|
			if c.layer != gh_layer
				hkeep_cdef[cdef.entityID] = true
				break
			end
		end	
	end
	
	#Building the lists of component to remove
	lst_gcomp.each do |comp|
		name = comp.get_attribute GHC__Dico, GHC__Name
		pair = @@hsh_pairs[name]
		next unless pair
		if comp.get_attribute(GHC__Dico, GHC__Type) == "Main"
			mdef = comp.definition
			gdef = pair.ghost_cdef
		elsif comp.get_attribute(GHC__Dico, GHC__Type) == "Ghost"
			gdef = comp.definition
			mdef = pair.main_cdef
		else
			next
		end
		unless (mdef && hkeep_cdef[mdef.entityID]) || (gdef && hkeep_cdef[gdef.entityID])
			ldel_comp.push comp
			hdel_cdef[gdef.entityID] = gdef if gdef
		end	
	end
	
	#Deleting the components and Ghost definitions
	@entities.erase_entities ldel_comp
	hdel_cdef.values.each { |cdef| cdef.entities.erase_entities cdef.entities.to_a if cdef.valid?}
	
end

#Cross reference the main and the ghost definitions
def associate_ghost_to_main(main_cdef, ghost_cdef)
	return unless main_cdef && ghost_cdef && main_cdef.valid? && ghost_cdef.valid?
	
	id = main_cdef.entityID
	@@hsh_maindef[id] = [] unless @@hsh_maindef[id]
	@@hsh_maindef[id].push ghost_cdef

	gname = GHC__PREFIX + main_cdef.name
	unless ghost_cdef.name == gname
		ghost_cdef.name = @definitions.unique_name(gname)
		ghost_cdef.description = T6[:DESC_GhostName, main_cdef.name]
	end	
end

#Create or update a pair Main / Ghost	
def register_pair(main_cdef, ghost_cdef)
	build_all_pairs
	
	#creating the pair
	pair = create_pair nil, main_cdef, ghost_cdef
	
	#Associating the definitions
	associate_ghost_to_main main_cdef, ghost_cdef
	
	#Assigning names to the Ghost definition	
	ghost_cdef.set_attribute GHC__Dico, GHC__Type, "Ghost"
	ghost_cdef.set_attribute GHC__Dico, GHC__Name, pair.name
	
	if SU_MAJOR_VERSION >= 7
		ghost_cdef.refresh_thumbnail
		main_cdef.refresh_thumbnail
	end	
	
	pair.name
end

#Mark a component with attributes, and set it on the ghost layer, hidden
def mark_component(comp, name, type)
	return unless comp && comp.valid?
	comp.hidden = true
	comp.layer = @layer_for_ghosts
	comp.set_attribute GHC__Dico, GHC__Name, name
	comp.set_attribute GHC__Dico, GHC__Type, type
end

#Check if the definition is Ghost
def is_ghost?(cdef)
	(cdef && cdef.get_attribute(GHC__Dico, GHC__Type) == "Ghost") ? true : false
end	

#Check if the definition is Main
def is_main?(cdef)
	(cdef && @@hsh_maindef[cdef.entityID]) ? true : false
end	

#Return the Ghost definition from the Main definition
def ghost_from_main(main_cdef, deepcheck=false)
	return nil unless is_main?(main_cdef)
	
	lgh = @@hsh_maindef[main_cdef.entityID]
	return unless lgh && lgh.length > 0
	
	#Repair (case of Undo)
	if lgh.find { |e| !e.valid? } && deepcheck
		build_all_pairs true
		return ghost_from_main(main_cdef)
	end
	
	lgh.each do |gdef|
		next unless gdef && gdef.valid? && gdef.instances.length > 0
		return gdef if gdef.instances.find { |comp| comp.layer == @layer_for_ghosts }
	end
	
	nil
end

#Return the Main definition from the Ghost definition
def main_from_ghost(gh_cdef)
	return nil unless is_ghost?(gh_cdef)
	
	#Getting the current main definition
	name = gh_cdef.get_attribute GHC__Dico, GHC__Name
	pair = get_pair name
	return nil unless pair
	mdef = pair.main_cdef
	(mdef && mdef.valid? ) ? mdef : nil
end

#Populate all ghost changes within other ghosts
def populate_ghosts
	@@hsh_pairs.each do |key, pair|
		main_cdef = pair.ghost_cdef
		next unless main_cdef && main_cdef.valid?
		gh_cdef = pair.ghost_cdef
		next unless gh_cdef && gh_cdef.valid?
		gh_cdef.entities.each do |e|
			next unless e.instance_of? Sketchup::ComponentInstance 
			cdef = e.definition
			if (gdef = ghost_from_main cdef)
				e.definition = gdef if gdef
			elsif (mdef = main_from_ghost cdef)
				gdef = ghost_from_main mdef
				e.definition = gdef if gdef
			end	
		end
	end
end

end	#class GhostCompAlgo
	
#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# Observer for Ghost instances opening in SU6 and SU7
#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
	
class GhostSelectionSpy #< Sketchup::SelectionObserver

# Method for SU6
# Seems to be the only callback received when selection is modified by click (but not when using contextual menu)
def onSelectionBulkChange(selection) 
	return if selection.length > 1
	comp = selection[0]
	
	#trapped a double click
	if comp == nil && @prev_comp && (Time.now.to_f - @prev_time) < 0.5
		ask_confirmation @prev_comp
		@prev_comp = nil
		@prev_time = Time.now.to_f
		return
	end	
	
	#Remember for possible double click, if selection is a ghost component
	if is_ghost?(comp)
		@prev_comp = comp
		@prev_time = Time.now.to_f
	else
		@prevcomp = nil
	end	
	
end

#Method for SU7, based on the event when entering a component edititing mode
#This is better, because it also catches entry via the contextual menu
def onActivePathChanged(*args)
	path = Sketchup.active_model.active_path
	return unless path
	comp = path.last
	return unless is_ghost?(comp)
	ask_confirmation comp
end

#Standard confirmation box for changes
#Return -1 to ignore changes, 0 to cancel and go back, +1 to save changes
def ask_confirmation(comp)
	html_text = "<b>" + T6[:WARNING_OpenGhost] + "</b>" + "\n" + T6[:T_WARNING_WhatToDo]	
	lbuttons = [:BUTTON_Edit, '^', :BUTTON_Close, '~', :BUTTON_Switch]
	title = T6[:TITLE_GhostComponent]
	UI.beep
	status = Traductor::WMsgBox.call html_text, nil, title, T6[lbuttons]
	
	case status
	when 1	#Close
		Sketchup.active_model.close_active
		
	when 2	#Switch
		Sketchup.active_model.close_active
		GhostCompAlgo.launcher :OPS_ToReal, nil, [comp]
			
	end	
end
	
def is_ghost?(comp)
	comp && comp.instance_of?(Sketchup::ComponentInstance) && comp.definition.get_attribute(GHC__Dico, GHC__Type) == "Ghost"
end

end	#class GhostSelectionSpy
	
end	#End Module GhostComp