=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed August 2009 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   GhsotLoader.rb
# Original Date	:   10 Aug 2009 - version 1.0
# Description	:   Loader for the Ghost Component script
# Menu Items	:   Tools --> submenu "Ghost Component"
# Toolbar		:   Name = "Ghost Component"
# Usage			:   See Tutorial and Quick Ref Card in PDF format
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module GhostComp

#---------------------------------------------------------------------------------------------------------------------------
# Overall Configuration for the Tools	
#---------------------------------------------------------------------------------------------------------------------------

MENU_PERSO = [:T_MenuTools_Fredo6Collection, "Tools", :T_MenuPlugins_Fredo6Collection, "Plugins"]

#---------------------------------------------------------------------------------------------------------------------------
#Constants for TOS  Modules (do not translate here, use Translation Dialog Box instead)	
#---------------------------------------------------------------------------------------------------------------------------

#Menu and Toolbar icons
T6[:GHC_MAKE_GHOST_Menu] = "Generate Ghost for the selection"
T6[:GHC_MAKE_GHOST_Tooltip] = "Create a ghost component"
GHC_MAKE_GHOST_Icon = "MakeGhost"

T6[:GHC_OPT_FREEZE_Menu] = "Option: CURRENT state"
T6[:GHC_OPT_FREEZE_Tooltip] = "Generate ghost from the CURRENT state of the components"
GHC_OPT_FREEZE_Icon = "OptFreeze"

T6[:GHC_OPT_BOX_Menu] = "Option: SIMPLE BOUNDING BOX"
T6[:GHC_OPT_BOX_Tooltip] = "Generate ghost as the bounding box of Top-level component"
GHC_OPT_BOX_Icon = "OptBox"

T6[:GHC_OPT_MULTIBOX_Menu] = "Option: CASCADING BOUNDING BOX"
T6[:GHC_OPT_MULTIBOX_Tooltip] = "Generate ghost as the bounding boxes of all components, recursively"
GHC_OPT_MULTIBOX_Icon = "OptMultiBox"

T6[:GHC_OPT_SIMPLE_Menu] = "Option: SIMPLIFY"
T6[:GHC_OPT_SIMPLE_Tooltip] = "Generate ghost as a simplified version of the components"
GHC_OPT_SIMPLE_Icon = "OptSimple"

T6[:GHC_OPT_WIREFRAME_Menu] = "Option: WIREFRAME"
T6[:GHC_OPT_WIREFRAME_Tooltip] = "Generate ghost as a wireframed version of the components"
GHC_OPT_WIREFRAME_Icon = "OptWireframe"

GHC_LIST_Options = [:GHC_OPT_MULTIBOX_, :GHC_OPT_BOX_, :GHC_OPT_FREEZE_, :GHC_OPT_WIREFRAME_, :GHC_OPT_SIMPLE_]

GHC_DEFAULT_Choice_Options = [ ['F', :GHC_OPT_FREEZE_Menu],
                               ['B', :GHC_OPT_BOX_Menu],
                               ['M', :GHC_OPT_MULTIBOX_Menu],
                               ['S', :GHC_OPT_SIMPLE_Menu],
                               ['W', :GHC_OPT_WIREFRAME_Menu] ]

T6[:GHC_DELETE_GHOST_Menu] = "Remove Ghost"
T6[:GHC_DELETE_GHOST_Tooltip] = "Remove the ghost of selected components"
GHC_DELETE_GHOST_Icon = "RemoveGhost"

T6[:GHC_CLEANUP_Menu] = "Clean up Ghosts"
T6[:GHC_CLEANUP_Tooltip] = "Clean up shadow versions of Ghosts (when components have been deleted)"
GHC_CLEANUP_Icon = "Cleanup"

T6[:GHC_TO_GHOST_Menu] = "Switch to Ghost"
T6[:GHC_TO_GHOST_Tooltip] = "Switch components to their Ghost version in the model"
GHC_TO_GHOST_Icon = "ToGhost"

T6[:GHC_TO_REAL_Menu] = "Switch to Real"
T6[:GHC_TO_REAL_Tooltip] = "Switch components to their Real version in the model"
GHC_TO_REAL_Icon = "ToReal"

T6[:GHC_TO_REAL_ALL_Menu] = "Switch to Real recursively"
T6[:GHC_TO_REAL_ALL_Tooltip] = "Switch recursively components to their Real version in the model"

#--------------------------------------------------------
# Default Parameters
#--------------------------------------------------------

T6[:GHC_DEFAULT_SECTION_General] = "Parameters at startup of Sketchup"
T6[:GHC_COLOR_RenderGhost] = "Color for rendering ghosts"
T6[:GHC_DEFAULT_RenderOption] = "Default option for creating Ghosts"
T6[:GHC_DEFAULT_NbSimple] = "Max number of faces to keep with option SIMPLIFY"
T6[:GHC_DEFAULT_UseObserver] = "Use observer (recommended for warning edition of ghost components)"
T6[:GHC_DEFAULT_AutoSwitch] = "Switch automatically to Ghost version after generation of Ghost"

@@ghost_option = nil

#--------------------------------------------------------
# Startup routine (called from LibFredo6)
#--------------------------------------------------------
	
#Create all commands with menus and toolbar icons
def GhostComp.startup	

	#Top menu
	MYPLUGIN.declare_topmenu nil, MENU_PERSO

	#Declaring command icons
	MYPLUGIN.declare_separator
	MYPLUGIN.declare_command(:GHC_MAKE_GHOST_) { GhostCompAlgo.launcher :OPS_CreateGhost }
	MYPLUGIN.declare_separator
	GHC_LIST_Options.each { |symb| MYPLUGIN.declare_command(symb) { GhostComp.set_option symb } }
	MYPLUGIN.declare_separator
	MYPLUGIN.declare_command(:GHC_DELETE_GHOST_) { GhostCompAlgo.launcher :OPS_DeleteGhost }
	MYPLUGIN.declare_command(:GHC_CLEANUP_) { GhostCompAlgo.new.launch_cleanup }
	MYPLUGIN.declare_separator
	MYPLUGIN.declare_command(:GHC_TO_GHOST_) { GhostCompAlgo.launcher :OPS_ToGhost }
	MYPLUGIN.declare_command(:GHC_TO_REAL_) { GhostCompAlgo.launcher :OPS_ToReal, :OPS_ToRealAll }
	MYPLUGIN.declare_command(:GHC_TO_REAL_ALL_) { GhostCompAlgo.launcher :OPS_ToReal ; GhostCompAlgo.launcher :OPS_ToRealAll }
	
	#Declaring the default parameter
	GhostComp.default_param
	
	#Startup of the Plugin
	MYPLUGIN.go	
	
	#Setting the default option for ghost generation
	GhostComp.set_default_option
	
	#Installing observer for change of model and reset parameters
	Sketchup.add_observer ModelSpy.new if MYDEFPARAM[:GHC_DEFAULT_UseObserver]	
end
	
#Declaring the Default Parameters	
def GhostComp.default_param
	dp = MYDEFPARAM
	dp.separator :GHC_DEFAULT_SECTION_General
	dp.declare :GHC_DEFAULT_RenderOption, 'B', "H:", T6[GHC_DEFAULT_Choice_Options]
	dp.declare :GHC_DEFAULT_NbSimple, 10, 'I:>1<150'
	dp.declare :GHC_COLOR_RenderGhost, "thistle", "K"	
	dp.declare :GHC_DEFAULT_UseObserver, true, "B"
	dp.declare :GHC_DEFAULT_AutoSwitch, false, "B"
	
	dp.add_notify_proc self.method("notify_default_param")
end
	
#Notify procedure when default parameters are changed	
def GhostComp.notify_default_param(symb, oldvalue)
	case symb
	when :GHC_COLOR_RenderGhost
		GhostCompAlgo.new.update_ghost_color
	end
end

#Get and Set the option for Ghost generation
def GhostComp.get_option ; @@ghost_option ; end

def GhostComp.set_default_option
	case MYDEFPARAM[:GHC_DEFAULT_RenderOption]
		when 'F' ; symb = :GHC_OPT_FREEZE_
		when 'S' ; symb = :GHC_OPT_SIMPLE_
		when 'W' ; symb = :GHC_OPT_WIREFRAME_
		when 'M' ; symb = :GHC_OPT_MULTIBOX_
		else ; symb = :GHC_OPT_BOX_
	end
	GhostComp.set_option symb
end
	
def GhostComp.set_option(option)	
	#Uncheck the current option if any	
	MYPLUGIN.set_button_state @@ghost_option, nil if @@ghost_option

	#setting the new option
	@@ghost_option = option
	MYPLUGIN.set_button_state @@ghost_option, MF_CHECKED
	topt = @@ghost_option.to_s + "Menu"
	tooltip = T6[:GHC_MAKE_GHOST_Tooltip] + " (" + T6[topt.intern] + ")"
	MYPLUGIN.change_button_tooltip :GHC_MAKE_GHOST_, tooltip
end

#API for external calls
def GhostComp.switch_to_ghost(lst_comp=nil) ; GhostCompAlgo.new.launch_operation :OPS_ToGhost, lst_comp, nil, true ; end
def GhostComp.switch_to_real(lst_comp=nil) ; GhostCompAlgo.new.launch_operation :OPS_ToReal, lst_comp, nil, true ; end

#--------------------------------------------------------
# Observer for model opening
# Install an observer for selection
#--------------------------------------------------------
	
class ModelSpy < Sketchup::AppObserver 

def initialize ; GhostCompAlgo.reset ; end
def onNewModel(model) ; GhostCompAlgo.reset ; end
def onOpenModel(model) ; GhostCompAlgo.reset ; end

end #class ModelSpy
	
end	#End Module GhostComp
